@@ -1,13 +1,20 @@
+# Copyright 2023 VMware, Inc. All Rights Reserved.
+# SPDX-License-Identifier: MIT
+
 from __future__ import annotations
 
 import asyncio
+import ssl
 from collections import defaultdict
 from dataclasses import dataclass
 from functools import partial
-from typing import Any, Optional, Union, cast
+from typing import Any, NoReturn, Optional, TypeVar
 
 from . import exceptions, schema, utils
-from .client import Client, ClientPool
+from .amqp import _MessageProtocol
+from .client import Addr, Client, ClientPool
+
+MessageT = TypeVar("MessageT", _MessageProtocol, bytes)
 
 
 @dataclass
@@ -20,9 +27,12 @@ class _Publisher:
 
 
 @dataclass
-class Message:
+class RawMessage(_MessageProtocol):
     data: bytes
-    id: Optional[int] = None
+    publishing_id: Optional[int] = None
+
+    def __bytes__(self) -> bytes:
+        return self.data
 
 
 class Producer:
@@ -31,34 +41,44 @@ class Producer:
         host: str,
         port: int = 5552,
         *,
-        vhost: str = '/',
+        ssl_context: Optional[ssl.SSLContext] = None,
+        vhost: str = "/",
         username: str,
         password: str,
         frame_max: int = 1 * 1024 * 1024,
         heartbeat: int = 60,
+        load_balancer_mode: bool = False,
+        max_retries: int = 20,
+        default_batch_publishing_delay: float = 0.2,
     ):
         self._pool = ClientPool(
             host,
             port,
+            ssl_context=ssl_context,
             vhost=vhost,
             username=username,
             password=password,
             frame_max=frame_max,
             heartbeat=heartbeat,
+            load_balancer_mode=load_balancer_mode,
+            max_retries=max_retries,
         )
         self._default_client: Optional[Client] = None
         self._clients: dict[str, Client] = {}
         self._publishers: dict[str, _Publisher] = {}
-        self._waiting_for_confirm: dict[
-            str,
-            dict[asyncio.Future[None], set[int]]
-        ] = defaultdict(dict)
+        self._waiting_for_confirm: dict[str, dict[asyncio.Future[None], set[int]]] = defaultdict(dict)
         self._lock = asyncio.Lock()
+        # dictionary [stream][list] of buffered messages to send asynchronously
+        self._buffered_messages: dict[str, list] = defaultdict(list)
+        self._buffered_messages_lock = asyncio.Lock()
+        self.task: asyncio.Task[NoReturn] | None = None
+        # Delay After sending the messages on _buffered_messages list
+        self._default_batch_publishing_delay = default_batch_publishing_delay
 
     @property
     def default_client(self) -> Client:
         if self._default_client is None:
-            raise ValueError('Producer is not started')
+            raise ValueError("Producer is not started")
         return self._default_client
 
     async def __aenter__(self) -> Producer:
@@ -72,6 +92,13 @@ class Producer:
         self._default_client = await self._pool.get()
 
     async def close(self) -> None:
+        # flush messages still in buffer
+        if self.task is not None:
+
+            for stream in self._buffered_messages:
+                await self._publish_buffered_messages(stream)
+            self.task.cancel()
+
         for publisher in self._publishers.values():
             await publisher.client.delete_publisher(publisher.id)
             publisher.client.remove_handler(schema.PublishConfirm, publisher.reference)
@@ -87,7 +114,7 @@ class Producer:
     async def _get_or_create_client(self, stream: str) -> Client:
         if stream not in self._clients:
             leader, _ = await self.default_client.query_leader_and_replicas(stream)
-            self._clients[stream] = await self._pool.get((leader.host, leader.port))
+            self._clients[stream] = await self._pool.get(Addr(leader.host, leader.port))
 
         return self._clients[stream]
 
@@ -100,14 +127,13 @@ class Producer:
             publisher = self._publishers[stream]
             if publisher_name is not None:
                 assert publisher.reference == publisher_name
-
             return publisher
 
         client = await self._get_or_create_client(stream)
 
         # We can have multiple publishers sharing same connection, so their ids must be distinct
         publisher_id = len([p for p in self._publishers.values() if p.client is client]) + 1
-        reference = publisher_name or f'{stream}_publisher_{publisher_id}'
+        reference = publisher_name or f"{stream}_publisher_{publisher_id}"
         publisher = self._publishers[stream] = _Publisher(
             id=publisher_id,
             stream=stream,
@@ -136,34 +162,46 @@ class Producer:
             partial(self._on_publish_error, publisher=publisher),
             name=publisher.reference,
         )
+
         return publisher
 
-    async def publish_batch(
+    async def send_batch(
         self,
         stream: str,
-        batch: Union[list[bytes], list[Message]],
+        batch: list[MessageT],
+        publisher_name: Optional[str] = None,
+    ) -> list[int]:
+
+        return await self._send_batch(stream, batch, sync=False, publisher_name=publisher_name)
+
+    async def _send_batch(
+        self,
+        stream: str,
+        batch: list[MessageT],
         sync: bool = True,
         publisher_name: Optional[str] = None,
     ) -> list[int]:
         if len(batch) == 0:
-            raise ValueError('Empty batch')
+            raise ValueError("Empty batch")
 
         async with self._lock:
-            publisher = await self._get_or_create_publisher(stream, publisher_name)
+            publisher = await self._get_or_create_publisher(stream, publisher_name=publisher_name)
 
         messages = []
+
         for item in batch:
-            if isinstance(item, Message):
-                message = schema.Message(
-                    publishing_id=item.id or publisher.sequence.next(),
-                    data=item.data,
+
+            msg = RawMessage(item) if isinstance(item, bytes) else item
+
+            if msg.publishing_id is None:
+                msg.publishing_id = publisher.sequence.next()
+
+            messages.append(
+                schema.Message(
+                    publishing_id=msg.publishing_id,
+                    data=bytes(msg),
                 )
-            else:
-                message = schema.Message(
-                    publishing_id=publisher.sequence.next(),
-                    data=item,
-                )
-            messages.append(message)
+            )
 
         await publisher.client.send_frame(
             schema.Publish(
@@ -171,6 +209,7 @@ class Producer:
                 messages=messages,
             ),
         )
+
         publishing_ids = [m.publishing_id for m in messages]
 
         if sync:
@@ -180,22 +219,63 @@ class Producer:
 
         return publishing_ids
 
-    async def publish(
+    async def send_wait(
         self,
         stream: str,
-        message: Union[Message, bytes],
-        sync: bool = True,
+        message: MessageT,
         publisher_name: Optional[str] = None,
     ) -> int:
-        messages = cast(Union[list[bytes], list[Message]], [message])
-        publishing_ids = await self.publish_batch(
+        publishing_ids = await self._send_batch(
             stream,
-            messages,
-            sync=sync,
+            [message],
+            sync=True,
             publisher_name=publisher_name,
         )
         return publishing_ids[0]
 
+    def _timer_completed(self, context):
+
+        if not context.cancelled():
+            if context.exception():
+                raise context.exception()
+
+        return 0
+
+    async def send(
+        self,
+        stream: str,
+        message: MessageT,
+        publisher_name: Optional[str] = None,
+    ):
+
+        # start the background thread to send buffered messages
+        if self.task is None:
+            self.task = asyncio.create_task(self._timer())
+            self.task.add_done_callback(self._timer_completed)
+
+        async with self._lock:
+            await self._get_or_create_publisher(stream, publisher_name=publisher_name)
+
+        async with self._buffered_messages_lock:
+            self._buffered_messages[stream].append(message)
+
+        await asyncio.sleep(0)
+
+    # After the timeout send the messages in _buffered_messages in batches
+    async def _timer(self):
+
+        while True:
+            await asyncio.sleep(self._default_batch_publishing_delay)
+            for stream in self._buffered_messages:
+                await self._publish_buffered_messages(stream)
+
+    async def _publish_buffered_messages(self, stream: str) -> None:
+
+        async with self._buffered_messages_lock:
+            if len(self._buffered_messages[stream]):
+                await self._send_batch(stream, self._buffered_messages[stream], sync=False)
+                self._buffered_messages[stream].clear()
+
     def _on_publish_confirm(self, frame: schema.PublishConfirm, publisher: _Publisher) -> None:
         if frame.publisher_id != publisher.id:
             return
@@ -221,10 +301,19 @@ class Producer:
                     fut.set_exception(exc)
                     del waiting[fut]
 
-    async def create_stream(self, stream: str, arguments: Optional[dict[str, Any]] = None) -> None:
-        await self.default_client.create_stream(stream, arguments)
+    async def create_stream(
+        self,
+        stream: str,
+        arguments: Optional[dict[str, Any]] = None,
+        exists_ok: bool = False,
+    ) -> None:
+        try:
+            await self.default_client.create_stream(stream, arguments)
+        except exceptions.StreamAlreadyExists:
+            if not exists_ok:
+                raise
 
-    async def delete_stream(self, stream: str) -> None:
+    async def delete_stream(self, stream: str, missing_ok: bool = False) -> None:
         if stream in self._publishers:
             publisher = self._publishers[stream]
             await publisher.client.delete_publisher(publisher.id)
@@ -232,7 +321,11 @@ class Producer:
             publisher.client.remove_handler(schema.PublishError, publisher.reference)
             del self._publishers[stream]
 
-        await self.default_client.delete_stream(stream)
+        try:
+            await self.default_client.delete_stream(stream)
+        except exceptions.StreamDoesNotExist:
+            if not missing_ok:
+                raise
 
     async def stream_exists(self, stream: str) -> bool:
         return await self.default_client.stream_exists(stream)
